# Developed by Aptus Engineering, Inc. <https://aptus.aero>
# See LICENSE.md file in project root directory

import json
from .packet import PacketType
from .data import DataPacket

class RequestPacket(DataPacket):
    '''
    Request a datapoint and type from/for a node.
    '''

    pType = PacketType.REQUEST_DATA
    pData = bytes()

    # Empty Constructor
    def __init__(self):
        super().__init__()


    # Construct from settings
    @staticmethod
    def constructFromRequest(node, image=False, data=False):
        p = RequestPacket()
        p.setData({
            "NODE": node,
            "IMAGE": image,
            "DATA": data
        })
        return p


# Developed by Aptus Engineering, Inc. <https://aptus.aero>
# See LICENSE.md file in project root directory

'''
Stream data packet.
Inherited from Packet class, implements:
    1. Constructor given stream data as a dict
    2. Parser to get stream data as a dict

Stream data payload format: JSON string (for flexibility)
Stream data may contain any variables and any data types supported by JSON.
'''

import json
from .packet import Packet, PacketType

class DataPacket(Packet):
    '''
    Stream data packet object. (see above)
    '''

    pType = PacketType.STREAM_DATA

    # Empty Constructor
    def __init__(self):
        super().__init__()


    # Construct data packet from dict
    @staticmethod
    def constructFromData(data):
        p = DataPacket()
        p.setData(data)
        return p


    # Set packet data from dict
    def setData(self, data):
        # Serialize dict to JSON string
        self.pData = bytes(json.dumps(data), 'utf-8') # UTF-8 encoded string


    # Set packet data from raw payload
    def setPayload(self, payload):
        self.pData = payload[1:] # First byte is a type identifier - ignore

    
    # Get telemetry data dict from object
    def getData(self):
        return json.loads(self.pData.decode())


# Developed by Aptus Engineering, Inc. <https://aptus.aero>
# See LICENSE.md file in project root directory

'''
Control command data packet.
Inherited from DataPacket class, implements:
    1. Constructor given control commands data as a dict
    2. Parser to get control commands data as a dict
'''

from .packet import PacketType
from .data import DataPacket

class ControlPacket(DataPacket):
    '''
    Control command data packet object. (see above)
    '''

    pType = PacketType.STREAM_DATA
    pData = bytes()

    # Empty Constructor
    def __init__(self):
        super().__init__()
    



# Developed by Aptus Engineering, Inc. <https://aptus.aero>
# See LICENSE.md file in project root directory

DATA_BUFFER_SIZE = 65536 #64kB

import os
import time
import socket
import asyncio
import websockets

from .protocol import *

# Load environment variables - includes connection config
from dotenv import load_dotenv
load_dotenv()

NODE_UUID = os.getenv('NODE_UUID', default="test")
ACCESS_KEY = os.getenv('ACCESS_KEY', default="testKey")
API_URL = os.getenv('API_URL', default="https://api.imperix.ai/")
STREAMER_URL = os.getenv('STREAMER_URI', default="ws://localhost:5000")


from .connection import connectionHandler


IMAGE_QUALITY_SIZE = [(360,240), (480,360), (640,480), (960,720)]


# NodeLink class definition
class NodeLink:
    '''
    A NodeLink object is used to interface with the Imperix Cloud.
    See the README.md file or read the complete documentation at:
    https://docs.imperix.ai
    '''

    socket = None

    # Image transmission feedback for auto-optimization
    imageQuality = 3 # Best quality
    imageLatency = 0
    imageFrameRate = 0


    # Constructor
    def __init__(self, missionUpdateCallback=None, manualControlCallback=None):
        '''
        Initialize link to Imperix Cloud for the node with environment config, and provided mission update callback and manual control callback.

        Parameters
        ----------
        @param missionUpdateCallback [func(mission)] - callback function to be called when node's mission is updated.
        @param manualControlCallback [func(control)] - callback function to be called when a manual control command is received.
        '''

        self.threadsActive = True

        # Set manual control callback thread
        self.manualControlCallback = manualControlCallback
        self.missionUpdateCallback = missionUpdateCallback


    # Destructor
    def __del__(self):
        self.disconnect() # Disconnect before deletion to kill threads


    # Connect
    async def connect(self):

        # Create socket to streamer
        self.socket = await websockets.connect(STREAMER_URL)

        # Send authorization packet
        p = AuthPacket.constructFromAuth(
            NODE_UUID,
            ACCESS_KEY,
            isNode=True
        )
        
        await p.transmit(self.socket)


    # Coroutine
    async def start(self):

        while True: # Try again if we loose connection

            if self.socket is None:
                await asyncio.sleep(1) # Wait to connect

            try:
                await connectionHandler(self)

            except Exception as e:
                print("NodeLink Socket Error:", e)
                await asyncio.sleep(1) # Wait and try again


    # Get mission
    async def fetchMission(self):
        
        pass


    # Mission update
    async def updateMission(self, activeWaypoint, missionStatus):
        '''
        Update mission state for display on Imperix Commander.

        Parameters
        ----------
        @param activeWaypoint [int] - active waypoint index, must be less than number of waypoints
        @param missionStatus [str] - mission status/error
        '''
        
        packet = MissionPacket()
        packet.setMissionStatus(
            activeWaypoint=activeWaypoint,
            missionStatus=missionStatus
        )

        await packet.transmit(self.socket)


    # Streaming data to the cloud
    # Telemetry
    async def transmitTelemetry(self, telemetry):
        '''
        Transmit dict with JSON-serializable telemetry parameters to Imperix Cloud.

        Parameters
        ----------
        @param telemetry [dict] - see README.md or documentation.
        '''

        packet = TelemetryPacket()

        packet.setData(telemetry)
        await packet.transmit(self.socket)

    
    # Image (array)
    async def transmitImage(self, image, timeStamp=None, feed='PRIMARY', optimize=True):
        '''
        Transmit image as numpy array to Imperix Cloud.

        Parameters
        ----------
        @param image [np.ndarray] - numpy array image (1 or 3 channels)
        @param timeStamp [datetime.datetime] - timestamp as a datetime object
        @param feed [str] - name of image/video feed
        @param optimize [bool] - auto-optimize framerate?
        '''

        # Set image transmission parameters to optimize for frame-rate
        if optimize:
            p = ImageData.constructFromImage(image, timeStamp, feed, resize=IMAGE_QUALITY_SIZE[self.imageQuality], compressionRatio=70)

        else:
            p = ImageData.constructFromImage(image, timeStamp, feed)
        
        await p.transmit(self.socket)

    
    # Image (binary)
    async def transmitImageBinary(self, image, timeStamp=None, feed='PRIMARY'):
        '''
        Transmit image as numpy array to Imperix Cloud.

        Parameters
        ----------
        @param image [bytes] - JPEG images as bytes object
        @param timeStamp [datetime.datetime] - timestamp as a datetime object
        @param feed [str] - name of image/video feed
        '''

        p = ImageData.constructFromBytes(image, timeStamp, feed)
        await p.transmit(self.socket)

    
    # Data stream
    async def transmitData(self, data):
        '''
        Transmit any JSON-serializable data to Imperix Cloud.

        Parameters
        ----------
        @param data [dict] - JSON-serializable data dictionary
        '''

        packet = DataPacket()
        packet.setData(data)
        await packet.transmit(self.socket)

    
    # Disconnect and close threads
    async def disconnect(self):
        '''
        Disconnect and close sockets, and kill threads.
        '''

        self.threadsActive = True

        # Close socket
        self.socket.close()


# Developed by Aptus Engineering, Inc. <https://aptus.aero>
# See LICENSE.md file in project root directory

'''
Telemetry data packet.
Inherited from Packet class, implements:
    1. Constructor given telemetry data as a dict
    2. Parser to get telemetry as a dict

Telemetry payload format: JSON string (for flexibility)

Telemetry payload object keys: (TODO: Keep updated!)

    1. Attitude
        ATT_ROLL (degrees)
        ATT_PITCH (degrees)
        ATT_HEADING (degrees)

    2. Location
        LOC_LATITUDE (degrees)
        LOC_LONGITUDE (degrees)
        LOC_ALTITUDE (meters)

    3. Velocities/Rates
        VEL_GROUNDSPEED (meters per second)
        VEL_AIRSPEED (meters per second)
        VEL_VERTICAL_SPEED (meters per second)
        VEL_ROLL_RATE (degrees per second)
        VEL_PITCH_RATE (degrees per second)
        VEL_YAW_RATE (degrees per second)

    4. Node Status
        STS_BATTERY (percentage 0-100)
        STS_SIGNAL (percentage 0-100)
'''

import json
from .packet import PacketType
from .data import DataPacket

class TelemetryPacket(DataPacket):
    '''
    Telemetry data packet object. (see above)
    '''

    pType = PacketType.TELEMETRY
    pData = bytes()

    # Empty Constructor
    def __init__(self):
        super().__init__()
    




# Developed by Aptus Engineering, Inc. <https://aptus.aero>
# See LICENSE.md file in project root directory

'''
The first packet received from a client must be an authorization
request packet. This is done through the stream instead of an API
request in order to capture the client address and save it to a
white list.

The authorization packet/header must include the following:
    IS_NODE: True if node, False if ui
    ID: UUID for node, or user
    KEY: Secret key associated with node, or user access token
'''

from .packet import PacketType
from .data import DataPacket

class AuthPacket(DataPacket):
    '''
    Authorization header packet object. (see above)
    '''
    pType = PacketType.AUTH_HEADER

    # Empty Constructor
    def __init__(self):
        super().__init__()
    

    # Construct from auth info
    @staticmethod
    def constructFromAuth(ident, key, isNode=False):
        p = AuthPacket()

        p.setData({
            "IS_NODE": isNode,
            "ID": ident,
            "KEY": key
        })

        return p


# Developed by Aptus Engineering, Inc. <https://aptus.aero>
# See LICENSE.md file in project root directory

'''
Protocol parser/generator for a given data point that was either
received (parser) or is to be transmitted (generator).
'''

import time
import asyncio

# Packet types
from enum import Enum
class PacketType(Enum):

    # Invalid/unset
    TYPE_UNSET =    b'\x00'

    # Data request from UI
    REQUEST_DATA =  b'\x01'

    # Authorization header
    AUTH_HEADER =   b'\x02'

    # Telemetry
    TELEMETRY =     b'\x03'

    # Stream data (any variable)
    STREAM_DATA =   b'\x04'

    # Manual control command
    MANUAL_CMD =    b'\x05'

    # Mission update
    MISSION =       b'\x06'

    # Image data
    IMAGE_FDBK =    b'\x07' # Stream signal feedback with 2-second average latency, and FPS
    IMAGE_START =   b'\x08' # First 4 payload bytes in UTC timestamp, then number of chunks, then data
    IMAGE_CHUNK =   b'\x09'


# Packet object (base class)
class Packet:
    '''
    A packet is either transmitted or received.

    A packet contains a data type id/field id (single byte),
    and the raw payload data.
    '''

    # Packet Type
    pType = PacketType.TYPE_UNSET
    pData = bytes()


    # Construct from type and data
    @staticmethod
    def constructFromData(packetType, packetData):
        '''
        Class constructor given packet type and data
        '''
        p = Packet()
        p.pType = packetType
        p.pData = packetData
        return p


    # Transmit packet
    async def transmit(self, socket):
        '''
        Transmit packet via socket.

        Parameters
        ----------
        @param socket [socket.socket] - connected socket object/reference
        '''
    
        await socket.send(self.pType.value + self.pData)


    # Parse packet - static method
    @staticmethod
    def parse(payload):
        '''
        Parse packet payload.

        Parameters
        ----------
        @param payload [bytes] - received bytes payload

        Returns
        -------
        @out packet [Packet] - parsed packet object
        '''

        if not isinstance(payload, bytes):
            raise Exception('Packet payload is not a bytes object.')

        if len(payload) < 1:
            raise Exception('Packet payload must have at least 1 byte.')

        # Create packet object to return
        try:
            packet = Packet.constructFromData(PacketType(bytes([payload[0]])), payload[1:])
            return packet

        except ValueError:
            raise Exception('Unknown packet type identifier.')

    
    # Receive packet and parse
    @staticmethod
    async def receive(socket):
        '''
        Receive and parse packet from socket.

        Parameters
        ----------
        @param socket [socket.socket] - connected socket object/reference

        Returns
        -------
        @out packet [Packet] - parsed packet object, or None if disconnected
        '''

        rx = await socket.recv()
        return Packet.parse(rx)







